# BASEPROMPT= ("The following is a description of a certain latent of text and a list of examples that contain the latent.\n"
#              "Description: \n"
#              "References to the Antichrist, the Apocalypse and conspiracy theories related to those topics. \n"
#              "Sentences:\n"
#              " by which he distinguishes Antichrist is, that he would rob God of his honour and take it to himself, he gives the leading latent which we ought \n"
#              " would be destroyed. The worlds economy would likely collapse as a result and could usher in a one world government movement. I wrote a small 6 page \n"
#              "3 begins. And the rise of Antichrist. Get ready with God that you would be found worthy to escape the horrors that will fall on earth. \n"
#              "Description: \n"
#              "Sentences containing digits forming a four-digit year, \n"
#              "Sentences:\n"
#              " 20, 2013 at 7:41 pm Martin Smith\nThis is all very interesting to me. I am related \n"
#              " of 2012. In other words, Italy's economy is now contracting at an at least 0.3% annualized run \n"
#              "end 2012 levels). In the first quarter of 2013, we expect revenue to be up slightly from the fourth quarter \n"
#              "Description: \n"
#              "Text related to banking and financial institutions \n"
#              "Sentences:\n"
#              ": He is on the Board of Directors with the Lumbee Bank jA course, at the present\ntime, he has a leave of absence with \n"
#              " refurbishing the Bank’s branches.\nBIP reached 400 thousand users in one year\nThe use of BIP has already doubled\nThe \n"
#              " the Federal Deposit Insurance Corp.\nMr. Ranzini would not say where University plans to open loan offices in the future.Beyond Michigan and the \n"
#              )
# BASEPROMPT= ("The following is a description of a certain latent of text and a list of examples that contain the latent.\n"
#              "Description: \n"
#              "References to the Antichrist, the Apocalypse and conspiracy theories related to those topics. \n"
#              "Sentences:\n"
#              " distinguishes Antichrist is, that he would rob \n"
#              " worlds economy would likely collapse as a result  \n"
#              "3 begins. And the rise of Antichrist. Get ready with  \n"
#              "Description: \n"
#              "Sentences containing digits forming a four-digit year, \n"
#              "Sentences:\n"
#              " 20, 2013 at 7:41 pm Martin Smith \n"
#              " of 2012. In other words, Italy's  \n"
#              "end 2012 levels). In the first quarter  \n"
#              "Description: \n"
#              "Text related to banking and financial institutions \n"
#              "Sentences:\n"
#              ": He is on the Board of Directors with the Lumbee Bank  \n"
#              " refurbishing the Bank’s branches. \n"
#              " the Federal Deposit Insurance Corp. \n"
#              )
BASEPROMPT = (
    "The following is a description of a certain latent of text and a list of examples that contain the latent.\n"
    "Description: \n"
    "References to the Antichrist, the Apocalypse and conspiracy theories related to those topics. \n"
    "Sentences: \n"
    " by which he distinguishes Antichrist is, that he would rob God of his honour and take it to himself, he gives the leading latent which we ought \n"
    "3 begins. And the rise of Antichrist. Get ready with  \n"
    " would be destroyed. The worlds economy would likely collapse as a result and could usher in a one world government movement. I wrote a small 6 page \n"
    "Description: \n"
    "Sentences containing digits forming a four-digit year, \n"
    "Sentences: \n"
    " 20, 2013 at 7:41 pm Martin Smith \n"
    " of 2012. In other words, Italy's  \n"
    "end 2012 levels). In the first quarter of 2013, we expect revenue to be up slightly from the fourth quarter \n"
    "Description: \n"
    "Text related to banking and financial institutions \n"
    "Sentences: \n"
    ": He is on the Board of Directors with the Lumbee Bank  \n"
    " refurbishing the Bank’s branches.\nBIP reached 400 thousand users in one year\nThe use of BIP has already doubled\nThe \n"
    " the Federal Deposit Insurance Corp. \n"
    "Description: \n"
    "Occurrences of the word 'The' at the beginning of sentences"
    "Sentences: \n"
    "The Smoking Tire hits the canyons with one of the fastest Audi's on the road \n"
    "The Chairman of the ABI \n"
    "The administrative center is the town of Koch. \n"
    "Description: \n"
    "Educational qualifications and degrees. \n"
    "Sentences: \n"
    "He also has a Masters in Business Administration degree from the National University. \n"
    "Robert J. McDonnell, B.S., Captain, USAF: Investigation of the High Angle of Fall of the Space Shuttle Challenger. \n"
    "He earned his PhD from the \n"
    "Description: \n"
    "Closing parenthesis and new line characters in programming languages."
    "Sentences: \n"
    ":COLOR_BACKGROUND);\ntft.add_text\n"
    "LOGICAL_X_POSITION(current_position.x)));\ntf"
    "Description: \n"
    "Hours and dates, mostly related to court operations. \n"
    "Sentences: \n"
    "The staff is available Monday through Friday from 8:30 \n"
    "when it was docketed in this court on January 23 \n"
    "Simpson mailed his complaint on or soon after January 16 \n"
    "Description: \n"
    "References to groups or communities, mostly women or men. \n"
    "Sentences: \n"
    "also retire from the little group of 50 beggar women with whom she had been working \n"
    "The men change out of jeans and sneakers \n"
    "Description: \n"
    "Terms and phrases related to 'rec' or 're' prefixes \n"
    "Sentences: \n"
    "1928 Recliner A recliner is a reclining chair. \n"
    "the memductance which is the reciprocal of memristance \n"
    " Recruits for this \n"
)
